/* dict.c - Abstract dictionary type, hash table variant
 * Copyright 2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: dict-hash.c,v 1.1.2.3 2001/07/09 15:01:34 entrope Exp $
 */

#include "config.h"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "common.h"
#include "dict.h"

#define RESIZE_LEVEL 0.85

struct dict_node {
    const char *key;
    void *data;
    struct dict_node *next;
};

struct dict {
    struct dict_node **table;
    unsigned int used, size;
    free_f free_keys, free_data;
};

#define POLYNOMIAL 0x04c11db7
static unsigned int crc_table[256];

static void
gen_crc_table(void)
{
    int i, j;
    unsigned int accum;
    for (i=0; i<256; i++) {
        accum = tolower(i) << 24;
        for (j=0; j<8; j++) {
            if (accum & 0x8000000) {
                accum = (accum << 1) ^ POLYNOMIAL;
            } else {
                accum = accum << 1;
            }
        }
        crc_table[i] = accum;
    }
}

__inline__ static unsigned int
dict_hash(const unsigned char *string)
{
    int accum = 0;
    while (*string) {
        accum = (accum << 8) ^ crc_table[((accum >> 24) ^ *string++) & 0xFF];
    }
    return accum;
}

dict_t
dict_new(void)
{
    static int crc_init;
    dict_t dict;
    if (!crc_init) {
        gen_crc_table();
        crc_init = 1;
    }
    dict = malloc(sizeof(struct dict));
    dict->used = 0;
    dict->size = 8;
    dict->table = malloc(dict->size*sizeof(dict->table[0]));
    memset(dict->table, 0, dict->size*sizeof(dict->table[0]));
    dict->free_keys = NULL;
    dict->free_data = NULL;
    return dict;
}

unsigned int
dict_size(dict_t dict)
{
    return dict->used;
}

void
dict_set_free_keys(dict_t dict, free_f free_keys)
{
    dict->free_keys = free_keys;
}

void
dict_set_free_data(dict_t dict, free_f free_data)
{
    dict->free_data = free_data;
}

const char *
dict_foreach(dict_t dict, dict_iterator_f it, void *extra)
{
    unsigned int i;
    struct dict_node *node;

    for (i=0; i<dict->size; i++) {
        for (node=dict->table[i]; node; node=node->next) {
            if (it(node->key, node->data, extra)) return node->key;
        }
    }
    return NULL;
}

static void
dict_resize(dict_t dict)
{
    unsigned int new_size, i, new_pos;
    struct dict_node **new_table, *node, *next;

    new_size = dict->size << 1;
    new_table = malloc(new_size*sizeof(new_table[0]));
    memset(new_table, 0, new_size*sizeof(new_table[0]));
    for (i=0; i<dict->size; i++) {
        for (node=dict->table[i]; node; node=next) {
            next = node->next;
            new_pos = dict_hash(node->key) % new_size;
            node->next = new_table[new_pos];
            new_table[new_pos] = node;
        }
    }
    dict->size = new_size;
    free(dict->table);
    dict->table = new_table;
}

void
dict_insert(dict_t dict, const char *key, void *data)
{
    struct dict_node *node, *last;
    unsigned int pos;

    pos = dict_hash(key) % dict->size;
    for (last = NULL, node = dict->table[pos];
         node;
         last = node, node = node->next) {
        if (!strcasecmp(node->key, key)) {
            if (dict->free_data) dict->free_data(node->data);
            if (dict->free_keys) dict->free_keys((void*)key);
            node->data = data;
            if (last) {
                last->next = node->next;
                node->next = dict->table[pos];
                dict->table[pos] = node;
            }
            return;
        }
    }
    node = malloc(sizeof(*node));
    node->key = key;
    node->data = data;
    node->next = dict->table[pos];
    dict->table[pos] = node;
    dict->used++;
    if (dict->used > dict->size * RESIZE_LEVEL) dict_resize(dict);
}

static void
dict_dispose_node(struct dict_node *node, free_f free_keys, free_f free_data)
{
    if (free_keys && node->key) free_keys((void*)node->key);
    if (free_data && node->data) free_data(node->data);
    free(node);
}

int
dict_remove2(dict_t dict, const char *key, int no_dispose)
{
    unsigned int pos;
    struct dict_node *last, *node;

    pos = dict_hash(key) % dict->size;
    for (last = NULL, node = dict->table[pos];
         node;
         last = node, node = node->next) {
        if (!strcasecmp(node->key, key)) break;
    }
    if (!node) return 0;
    if (last) {
        last->next = node->next;
    } else {
        dict->table[pos] = node->next;
    }
    if (no_dispose) {
        free(node);
    } else {
        dict_dispose_node(node, dict->free_keys, dict->free_data);
    }
    dict->used--;
    return 1;
}

void *
dict_find(dict_t dict, const char *key, int *found)
{
    unsigned int pos;
    struct dict_node *last, *node;

    pos = dict_hash(key) % dict->size;
    for (last = NULL, node = dict->table[pos];
         node;
         last = node, node = node->next) {
        if (!strcasecmp(node->key, key)) break;
    }
    if (node) {
        if (last) {
            last->next = node->next;
            node->next = dict->table[pos];
            dict->table[pos] = node;
        }
        if (found) *found = 1;
        return node->data;
    } else {
        if (found) *found = 0;
        return NULL;
    }
}

void
dict_delete(dict_t dict)
{
    unsigned int i;
    struct dict_node *node, *next;

    for (i=0; i<dict->size; i++) {
        for (node=dict->table[i]; node; node=next) {
            next = node->next;
            dict_dispose_node(node, dict->free_keys, dict->free_data);
        }
    }
    free(dict->table);
    free(dict);
}

char *
dict_sanity_check(dict_t dict)
{
    unsigned int i;
    struct dict_node *node;
    char buffer[128];

    for (i=0; i<dict->size; i++) {
        for (node=dict->table[i]; node; node=node->next) {
            if ((dict_hash(node->key) % dict->size) != i) {
                snprintf(buffer, sizeof(buffer), "Bucket %d contained out-of-place string '%s'", i, node->key);
                return strdup(buffer);
            }
        }
    }
    return NULL;
}
